//-*-c++-*------------------------------------------------------------
//
// File name : bioArithDerivative.cc
// Author :    Michel Bierlaire
// Date :      Tue May 11 06:07:21 2010
//
//--------------------------------------------------------------------

#include <sstream>

#include "patMath.h"
#include "bioParameters.h"
#include "patDisplay.h"
#include "patErrNullPointer.h"
#include "patErrMiscError.h"
#include "bioArithDerivative.h"
#include "bioArithConstant.h"
#include "bioArithLiteral.h"
#include "bioArithMult.h"
#include "bioArithBinaryPlus.h"
#include "bioArithDivide.h"
#include "bioLiteralRepository.h"
#include "bioArithNotEqual.h"

bioArithDerivative::bioArithDerivative(bioExpression* par,
				       bioExpression* left,
				       patString aLiteralName, patError*& err)
  : bioExpression(par,left,NULL), literalName(aLiteralName)
{
  DEBUG_MESSAGE("Create derivative object") ;
  literalId = bioLiteralRepository::the()->getLiteralId(literalName,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
  }

  if (left == NULL) {
    err = new patErrNullPointer("biopExpression") ;
    WARNING(err->describe()) ;
    return ;
  }
  theDerivative = left->getDerivative(literalId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  
//   DEBUG_MESSAGE("Derive the following expression wrt to " << literalName << " [" << literalId << "]") ;
//   DEBUG_MESSAGE("Expression: " << *leftChild) ;
//   DEBUG_MESSAGE("Result: " << *theDerivative) ;
}


bioArithDerivative::~bioArithDerivative() {}

bioExpression::bioOperatorType bioArithDerivative::getOperatorType() const {
  return bioExpression::DERIVATIVE_OP ;
}

patString bioArithDerivative::getOperatorName() const {
  return ("Derivative") ;
}

patReal bioArithDerivative::getValue(patError*& err) const  {
  err = new patErrMiscError("Not implemented.") ;
  WARNING(err->describe()) ;
  return patReal() ;
}


bioExpression* bioArithDerivative::getDerivative(patULong aLiteralId, 
					       patError*& err) const {

  if (leftChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    WARNING(getInfo()) ;
    return NULL;
  }
  patString derivName = bioLiteralRepository::the()->getName(aLiteralId,err) ;

  if (aLiteralId == literalId) {
    stringstream str ;
    str << "Cannot compute the derivative with respect to " << derivName << " of an integral over " << literalName ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return NULL ;
  }

  bioExpression* result = theDerivative->getDerivative(aLiteralId,err) ;

  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  return result ;
}


bioArithDerivative* bioArithDerivative::getDeepCopy(patError*& err) {
  bioExpression* leftClone(NULL) ;
  if (leftChild != NULL) {
    leftClone = leftChild->getDeepCopy(err) ;
  }
  bioArithDerivative* theNode = 
    new bioArithDerivative(NULL,leftClone,literalName,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  leftClone->setParent(theNode) ;
  return theNode ;
}

patString bioArithDerivative::getSimpleCppCode(patError*& err) {
  err = new patErrMiscError("No simple c++ code for a derivative") ;
  WARNING(err->describe());
  return patString() ;
}


patString bioArithDerivative::getCppCode(patString prefix, patError*& err) {
  if (codeMustBeGenerated) {
    if (leftChild == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return patString();
    }
    stringstream str ;
    if (theDerivative->isSimple()) {
      str << getCppResultDeclaration(prefix) << " = " << theDerivative->getSimpleCppCode(err)  << ";" << endl ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return patString() ;
      }
    }
    else {
      str <<  theDerivative->getCppCode(prefix,err) << endl ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return patString() ;
      }      
      str << getCppResultDeclaration(prefix) << " = " << theDerivative->getCppResult(prefix) << ";" ; 
    }
    

    codeMustBeGenerated = patFALSE ;
    return patString(str.str()) ;
  }
  else {
    return patString() ;
  }
}

patBoolean bioArithDerivative::isStructurallyZero() const {
  return (leftChild->isStructurallyZero()) ;
}

patString bioArithDerivative::getExpressionString() const {
  stringstream str ;
  str << 'Deriv' ;
  if (leftChild != NULL) {
    str << '{' << leftChild->getExpressionString() << '}' ;
  }
  return patString(str.str()) ;

}

void bioArithDerivative::generateGslFunctions(ostream& str_h,patString prefix,patError*& err) {
  if (leftChild != NULL) {
    leftChild->generateGslFunctions(str_h,prefix,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
  }
  if (theDerivative != NULL) {
    theDerivative->generateGslFunctions(str_h,prefix,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
  }


}

