//-*-c++-*------------------------------------------------------------
//
// File name : bioArithDivide.cc
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Tue Apr 28 20:58:50 2009
//
//--------------------------------------------------------------------

#include <sstream>

#include "patDisplay.h"
#include "patErrNullPointer.h"
#include "patErrMiscError.h"
#include "bioLiteralRepository.h"
#include "bioArithDivide.h"
#include "bioArithMult.h"
#include "bioArithUnaryMinus.h"
#include "bioArithBinaryMinus.h"
#include "bioArithPower.h"
#include "bioArithConstant.h"
#include "bioArithLiteral.h"
#include "bioArithSequence.h"

bioArithDivide::bioArithDivide(bioExpression* par,
                               bioExpression* left, 
                               bioExpression* right) : bioExpression(par,left,right)
{
}

bioArithDivide::~bioArithDivide() {}

bioExpression::bioOperatorType bioArithDivide::getOperatorType() const {
  return bioExpression::BINARY_OP ;
}

patString bioArithDivide::getOperatorName() const {
  return ("/") ;
}

patReal bioArithDivide::getValue(patError*& err) const  {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patReal();
  }
  
  patReal left = leftChild->getValue(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }
  if (left == 0.0) {
    return 0.0 ;
  }

  patReal right = rightChild->getValue(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }
  if (right == 0.0) {
    err = new patErrMiscError("Divide by zero error") ;
    WARNING(err->describe()) ;
    return patReal();
  }

  patReal result = left / right ;

  return result ;
}

bioExpression* bioArithDivide::getDerivative(patULong aLiteralId, 
					     patError*& err) const {

  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return NULL;
  }
  
  bioExpression* leftValue = leftChild->getDeepCopy(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  bioExpression* rightValue = rightChild->getDeepCopy(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  bioExpression* leftDeriv = leftChild->getDerivative(aLiteralId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  bioExpression* rightDeriv = rightChild->getDerivative(aLiteralId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  if (leftChild->dependsOf(aLiteralId)) {
    if (rightChild->dependsOf(aLiteralId)) {
      patULong theId = bioLiteralRepository::the()->addIntermediateValue(err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
      }
      bioArithLiteral* theSimplifiedExpression =  new bioArithLiteral(NULL,theId) ;
      // Left variable, right variable
      bioExpression* l = new bioArithMult(NULL,theSimplifiedExpression,leftDeriv) ;
      bioExpression* r = new bioArithMult(NULL,leftValue,rightDeriv) ;
      bioExpression* n = new bioArithBinaryMinus(NULL,l,r) ;
      bioExpression* d = new bioArithMult(NULL,theSimplifiedExpression,theSimplifiedExpression) ;
      bioExpression* ratio = new bioArithDivide(NULL,n,d) ;
      bioArithSequence* result = new bioArithSequence(NULL,rightValue,ratio,theId,err) ; 
      if (err != NULL) {
	WARNING(err->describe()) ;
	return NULL;
      }
      return result ;
    }
    else {
      // Left variable, right constant
      bioExpression* l = new bioArithDivide(NULL,leftDeriv,rightValue) ;
      return l ;
    }
  }
  else {
    if (rightChild->dependsOf(aLiteralId)) {
      //Left constant, right variable
      bioExpression* r = new bioArithMult(NULL,leftValue,rightDeriv) ;
      bioExpression* n = new bioArithUnaryMinus(NULL,r) ;
      bioExpression* d = new bioArithMult(NULL,rightValue,rightValue) ;
      bioExpression* result = new bioArithDivide(NULL,n,d) ;
      return result ;
    }
    else {
      //Left constant, right constant
      bioExpression* result = new bioArithConstant(NULL,0) ;
      return result ;
    }
  }

  
  //  patReal result = (rightValue * leftDeriv - leftValue * rightDeriv) / 
  //                 (rightValue * rightValue) ;

}

bioArithDivide* bioArithDivide::getDeepCopy(patError*& err) {
  bioExpression* leftClone(NULL) ;
  bioExpression* rightClone(NULL) ;
  if (leftChild != NULL) {
    leftClone = leftChild->getDeepCopy(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    if (leftClone == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return NULL ;
    }
  }
  if (rightChild != NULL) {
    rightClone = rightChild->getDeepCopy(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    if (rightClone == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return NULL ;
    }
  }

  bioArithDivide* theNode = 
    new bioArithDivide(NULL,leftClone,rightClone) ;
  leftClone->setParent(theNode) ;
  rightClone->setParent(theNode) ;
  return theNode ;
}

patString bioArithDivide::getCppCode(patString prefix, patError*& err) {
  if (codeMustBeGenerated) {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString();
  }
  
  patBoolean lSimple = leftChild->isSimple() ;
  patString vLeft ;
  if (lSimple) {
    vLeft = leftChild->getSimpleCppCode(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  else {
    vLeft = leftChild->getCppCode(prefix, err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  patBoolean rSimple = rightChild->isSimple() ;
  patString vRight ;
  if (rSimple) {
    vRight = rightChild->getSimpleCppCode(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  else {
    vRight = rightChild->getCppCode(prefix, err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  stringstream str ;
  str << getCppResultDeclaration(prefix) << ";" ;
  if (!lSimple) {
    str << vLeft << endl ;
  }
  if (!rSimple) {
    str << vRight << endl ;
  }
  if (!lSimple) {
    str << "if (" <<leftChild->getCppResult(prefix) << "== 0) {" << endl ;
    str << "    " << getCppResult(prefix) << "= 0 ;" << endl ;
    str << "}" << endl ;
  }
  else {
    str << "if ((" << vLeft << ") == 0) {" << endl ;
    str << "    " << getCppResult(prefix) << "= 0 ;" << endl ;
    str << "}" << endl ;
  }
  str << " else {" << endl ;
  str << getCppResult(prefix) << "= (" ;
  if (lSimple) {
    str << vLeft ;
  }
  else {
    str << leftChild->getCppResult(prefix) ;
  }
  str << " ) / ( " ;
  if (rSimple) {
    str << vRight ;
  }
  else {
    str << rightChild->getCppResult(prefix) ;
  }
  str << ");" << endl ;
  str << "}" << endl ;
  str << "    if (!finite(" << getCppResult(prefix) << ")) {" << endl ;
  str << "      " << getCppResult(prefix) << " =  patMaxReal ;" << endl ;
  str << "    }" << endl ;
  //  str << "DEBUG_MESSAGE(\"" << getCppResult(prefix) << "= \" << " << getCppResult(prefix) << ");" << endl ;
  codeMustBeGenerated = patFALSE ;
  return patString(str.str()) ;
  }
  else {
    return patString() ;
  }
}


patString bioArithDivide::getSimpleCppCode(patError*& err) {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString();
  }
  
  patString vLeft = leftChild->getSimpleCppCode(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  patString vRight = rightChild->getSimpleCppCode(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }

  if (vLeft == "0.0") {
    return patString("0.0") ;
  }

  stringstream str ;
  str << "(((" << vLeft << ")==0)?0:((" << vLeft << ")/(" << vRight << ")))"  ;
  return patString(str.str()) ;
}



patBoolean bioArithDivide::isSimple() const {
  return patFALSE ;
}

patBoolean bioArithDivide::isStructurallyZero() const {
  return (leftChild->isStructurallyZero()) ;
}

patString bioArithDivide::getExpressionString() const {
  stringstream str ;
  str << '/' ;
  if (leftChild != NULL) {
    str << '{' << leftChild->getExpressionString() << '}' ;
  }
  if (rightChild != NULL) {
    str << '{' << rightChild->getExpressionString() << '}' ;
  }
  return patString(str.str()) ;

}
