//-*-c++-*------------------------------------------------------------
//
// File name : bioArithElem.cc
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Tue May  5 10:56:09 2009
//
//--------------------------------------------------------------------

#include <sstream>

#include "patDisplay.h"
#include "patErrNullPointer.h"
#include "patErrMiscError.h"

#include "bioParameters.h"
#include "bioArithSequence.h"
#include "bioArithElem.h"
#include "bioArithConstant.h"
#include "bioLiteralRepository.h"

bioArithElem::bioArithElem(bioExpression* par,
                           bioExpression* left,
			   map<patULong,bioExpression*> aDict,
			   bioExpression* defExp) : 
  bioExpression(par,left,NULL),
  theDictionary(aDict),
  defaultExpression(defExp) {
  for (map<patULong,bioExpression*>::iterator i = theDictionary.begin() ;
       i != theDictionary.end() ;
       ++i) {
    i->second->setParent(this) ;
  }
}

bioArithElem::~bioArithElem() {}

bioExpression::bioOperatorType bioArithElem::getOperatorType() const {
  return bioExpression::DICTIONARY_OP ;
}

patString bioArithElem::getOperatorName() const {
  return ("elem") ;
}


patReal bioArithElem::getValue(patError*& err) const  {
  patReal index = leftChild->getValue(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal() ;
  }
  map<patULong,bioExpression*>::const_iterator found = theDictionary.find(patULong(index)) ;
  if (found == theDictionary.end()) {
    stringstream str ;
    str << "Index " << index << " not valid in expression " << *this ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return patReal() ;
  }
  patReal result = found->second->getValue(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal() ;
  }
  return result ;
}

bioExpression* bioArithElem::getDerivative(patULong aLiteralId, patError*& err) const {
  map<patULong,bioExpression*> theDerivatives ;
  for (map<patULong,bioExpression*>::const_iterator i = theDictionary.begin() ;
       i != theDictionary.end() ;
       ++i) {
    bioExpression* theDeriv = i->second->getDerivative(aLiteralId,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL;
    }
    theDerivatives[i->first] = theDeriv ;
  }

  bioExpression* derivDefault = defaultExpression->getDerivative(aLiteralId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }

  bioExpression* result = new bioArithElem(NULL,leftChild,theDerivatives,derivDefault) ;
  return result ;
}

bioArithElem* bioArithElem::getDeepCopy(patError*& err) {
  bioExpression* leftClone(NULL) ;
  bioExpression* rightClone(NULL) ;
  if (leftChild != NULL) {
    leftClone = leftChild->getDeepCopy(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
  }
  if (rightChild != NULL) { // Must be NULL
    rightClone = rightChild->getDeepCopy(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
  }

  map<patULong,bioExpression*> newDictionary ;

  for (map<patULong,bioExpression*>::iterator e = theDictionary.begin() ;
       e != theDictionary.end() ;
       ++e) {
    bioExpression* clone = e->second->getDeepCopy(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    newDictionary[e->first] = clone ;
  }

  bioExpression* defaultClone = defaultExpression->getDeepCopy(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }

  bioArithElem* theNode = 
    new bioArithElem(NULL,leftClone,newDictionary,defaultClone) ;
  leftClone->setParent(theNode) ;
  return theNode ;
}


patString bioArithElem::getCppCode(patString prefix, patError*& err) {
  if (codeMustBeGenerated) {
    codeMustBeGenerated = patFALSE ;
    if (leftChild == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return patString();
    }
  
    patBoolean defSimple = defaultExpression->isSimple() ;

    patString defResult ;
    if (defSimple) {
      defResult = defaultExpression->getSimpleCppCode(err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return patString();
      }  
    }
    else {
      defResult = defaultExpression->getCppCode(prefix,err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return patString();
      }  
    }




    patBoolean lSimple = leftChild->isSimple() ;
  
    patString result ;
    if (lSimple) {
      result = leftChild->getSimpleCppCode(err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return patString();
      }  
    }
    else {
      result = leftChild->getCppCode(prefix,err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return patString();
      }  
    }

    //  DEBUG_MESSAGE("Cpp code for " << *leftChild << " is " << result) ;
    stringstream str ;
    if (!lSimple) {
      str << result << endl ;
    }

    if (!defSimple) {
      str << defResult << endl ;
    }


    str << getCppResultDeclaration(prefix) << " ; " << endl ;
    str << "switch (patULong(" ;
    if (lSimple) {
      str << result ;
    }
    else {
      str << leftChild->getCppResult(prefix) ;
    }
    str << ")) {" << endl ;
    for (map<patULong,bioExpression*>::iterator i =  theDictionary.begin() ;
	 i != theDictionary.end() ;
	 ++i) {
      str << "case " << i->first << ":" << endl ;
      str << "  { " << endl ;
      if (!i->second->isSimple()) {
	str << i->second->getCppCode(prefix, err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return patString();
	}
      }
      str << getCppResult(prefix) << " = " ;
      if (i->second->isSimple()) {
	str <<  i->second->getSimpleCppCode(err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return patString();
	}
      }
      else {
	str << i->second->getCppResult(prefix) ;
      }
      str << " ; " << endl ; 
      str << " } " << endl ; 
      str << "break ; " << endl ;
    }
    str << " default: {" << endl ;

    
    str << "    stringstream errstr ;" << endl ;
    str << "    errstr << \"Illegal value in Elem dictionary: "<<*leftChild<<"  = \"" << endl ;
    if (lSimple) {
      str << " << " << result ;
    }
    else {
      str << " << " << leftChild->getCppResult(prefix) ;
    }
    str << "; " << endl ;
    str << "    errstr << \". Dictionary expects one of the following values: \"" << endl ;
    for (map<patULong,bioExpression*>::iterator i =  theDictionary.begin() ;
	 i != theDictionary.end() ;
	 ++i) {
      if (i != theDictionary.begin()) {
	str << " << \",\" "  ;
      }
      str << " << " << i->first ;
    }
    str << ";" << endl ;

    str << " errstr << \"Default expression is used: " << *defaultExpression << " = \" " << endl ;

    //    str << "    err = new patErrMiscError(errstr.str()) ;" << endl ;
    //  str << "    WARNING(err->describe()) ;" << endl ;
    // str << "    return NULL ;" << endl ;

    if (defSimple) {
      str << " << " << defResult ;
    }
    else {
      str << " << " << defaultExpression->getCppResult(prefix) ;
    }
    str << "; " << endl ;

    long warning = bioParameters::the()->getValueInt("warnsForIllegalElements",err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
    if (warning != 0) {
      str << " cerr << patString(errstr.str()) << endl ; " << endl ;
    }

    str << getCppResult(prefix) << " = " ;
    if (defSimple) {
      
      str << defResult << " ;" << endl ;
    }
    else {
      str << defaultExpression->getCppResult(prefix) << ";" << endl  ;
    }

    str << "}" << endl;

    str << "}" << endl;

    variableWhereStored = getCppResult(prefix) ;
    //    str << "DEBUG_MESSAGE(\"" << getCppResult(prefix) << "= \" << " << getCppResult(prefix) << ");" << endl ;
  codeMustBeGenerated = patFALSE ;
    return patString(str.str()) ;
  }
  else {
    return patString() ;
  }
}




patString bioArithElem::getExpression(patError*& err) const {
  stringstream str ;
  str << "Elem(" ;
  for (map<patULong,bioExpression*>::const_iterator i = theDictionary.begin() ;
       i != theDictionary.end();
       ++i) {
    if (i != theDictionary.begin()) {
      str << "," ;
    }
    patString theExpr = i->second->getExpression(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
    str << "{" << i->first << ": " << theExpr << "}" ;
  }
  patString theIndex = leftChild->getExpression(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  patString theDefault = defaultExpression->getExpression(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  str << "[" << theIndex << "][" << theDefault << "]" ;
  return patString(str.str()) ;
  
}

patBoolean bioArithElem::isSimple() const {
  return (codeMustBeGenerated)?patFALSE:patTRUE ;
}

patString bioArithElem::getSimpleCppCode(patError*& err) {
  return variableWhereStored ;
}

patBoolean bioArithElem::dependsOf(patULong aLiteralId) const {
  for (map<patULong,bioExpression*>::const_iterator i = theDictionary.begin() ;
       i != theDictionary.end() ;
       ++i) {
    if (i->second->dependsOf(aLiteralId)) {
      return patTRUE ;
    }
  }
  if (defaultExpression->dependsOf(aLiteralId)) {
    return patTRUE ;
  }
  return patFALSE ;
}

void bioArithElem::forceCodeGeneration() {
  codeMustBeGenerated = patTRUE ;
  leftChild->forceCodeGeneration() ;
  defaultExpression->forceCodeGeneration() ;
  bioLiteralRepository::the()->resetAllFlags() ;
  for (map<patULong,bioExpression*>::const_iterator i = theDictionary.begin() ;
       i != theDictionary.end() ;
       ++i) {
    i->second->forceCodeGeneration() ;
  }
}

void bioArithElem::simplifyZeros(patError*& err) {
  if (leftChild != NULL) {
    if (leftChild->isStructurallyZero()) {
      //      DELETE_PTR(leftChild) ;
      leftChild = new bioArithConstant(this,0) ;
    }
    else {
      leftChild->simplifyZeros(err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
    }
  }
  if (defaultExpression != NULL) {
    if (defaultExpression->isStructurallyZero()) {
      //      DELETE_PTR(leftChild) ;
      defaultExpression = new bioArithConstant(this,0) ;
    }
    else {
      defaultExpression->simplifyZeros(err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
    }
  }
  for (map<patULong,bioExpression*>::const_iterator i = theDictionary.begin() ;
       i != theDictionary.end() ;
       ++i) {
    if (i->second->isStructurallyZero()) {
      theDictionary[i->first] = new bioArithConstant(NULL,0) ;
    }
    else {
      i->second->simplifyZeros(err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
    }
  }
  
}

patULong bioArithElem::getNumberOfOperations() const {
  patULong result = 0 ;
  if (leftChild != NULL) {
    result += leftChild->getNumberOfOperations() ;
  }
  if (defaultExpression != NULL) {
    result += defaultExpression->getNumberOfOperations() ;
  }
  // Although it is not really correct, we consider the sum of all operations
  for (map<patULong,bioExpression*>::const_iterator i = theDictionary.begin() ;
       i != theDictionary.end() ;
       ++i) {
      result += i->second->getNumberOfOperations() ;
  }
  return (result) ;
}

patULong bioArithElem::lastIndex() const {
  return theDictionary.rbegin()->first ;
}

patULong bioArithElem::countExpressions(map<patString,patULong>* theExpressions, 
					map<patString,list<patULong> >* theIds, 
					map<patString,list<bioExpression*> >* thePointers,
					patError*& err)   {
  patULong result(0) ;
  if (getNumberOfOperations() > 0) {
    stringstream str ;
    str << getExpressionString() ;
    (*theExpressions)[patString(str.str())] += 1 ;
    (*theIds)[patString(str.str())].push_back(getId()) ;
    (*thePointers)[patString(str.str())].push_back(this) ;
    result = getNumberOfOperations() ;
  }
  for (map<patULong,bioExpression*>::const_iterator i = theDictionary.begin() ;
       i != theDictionary.end() ;
       ++i) {
    i->second->countExpressions(theExpressions,theIds,thePointers,err) ;
  }
  return result ;
}

bioExpression* bioArithElem::getExpressionFromId(patULong theId) {
  if (theId == getId()) {
    return this ;
  }
  bioExpression* result(NULL) ;
  if (leftChild != NULL) {
    result = leftChild->getExpressionFromId(theId) ;
    if (result != NULL) {
      return result ;
    }
  }
  for (map<patULong,bioExpression*>::const_iterator i = theDictionary.begin() ;
       i != theDictionary.end() ;
       ++i) {
    result = i->second->getExpressionFromId(theId) ;
    if (result != NULL) {
      return result ;
    }
  }
  if (defaultExpression != NULL) {
    result = defaultExpression->getExpressionFromId(theId) ;
    if (result != NULL) {
      return result ;
    }
  }
  return NULL ;
  
}

void bioArithElem::replaceExpression(bioExpression* theExpression, 
				     bioExpression* anotherExpression, 
				     patError*& err) {
  
  for (map<patULong,bioExpression*>::iterator i = theDictionary.begin() ;
       i != theDictionary.end() ;
       ++i) {
    if (*(i->second) == *theExpression) {
      //DELETE_PTR(i->second) ;
      i->second = anotherExpression->getDeepCopy(err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
      i->second->setParent(this) ;
    }
    else {
      i->second->replaceExpression(theExpression,anotherExpression,err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
    }

  }
}




patString bioArithElem::getExpressionString() const {
  stringstream str ;
  str << '$D' ;
  if (leftChild != NULL) {
    str << '[' << leftChild->getExpressionString() << ']' ;
  }
  if (defaultExpression != NULL) {
    str << '[' << defaultExpression->getExpressionString() << ']' ;
  }
  str << '[' ;
  for (map<patULong,bioExpression*>::const_iterator i = theDictionary.begin() ;
       i != theDictionary.end() ;
       ++i) {
    //    DEBUG_MESSAGE("BEFORE: " << *i->second) ;
    str << "{" << i->second->getExpressionString() << "}" ;
  }  
  str << ']' ;
  return patString(str.str()) ;
}

void bioArithElem::getExpressionsToSimplify(vector<bioExpression*>* theExpr,
					    patError*& err) {
  for (map<patULong,bioExpression*>::iterator i = theDictionary.begin() ;
       i != theDictionary.end() ;
       ++i) {
    i->second->getExpressionsToSimplify(theExpr,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
  }
  defaultExpression->getExpressionsToSimplify(theExpr,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
}

patBoolean bioArithElem::containsAnIterator() const {
  for (map<patULong,bioExpression*>::const_iterator i = theDictionary.begin() ;
       i != theDictionary.end() ;
       ++i) {
    if (i->second->containsAnIterator()) {
      return patTRUE ;
    }
  }

  if (defaultExpression->containsAnIterator()) {
    return patTRUE ;
  }
  return patFALSE ;
}

patBoolean bioArithElem::containsAnIntegral() const {
  for (map<patULong,bioExpression*>::const_iterator i = theDictionary.begin() ;
       i != theDictionary.end() ;
       ++i) {
    if (i->second->containsAnIntegral()) {
      return patTRUE ;
    }
  }
  if (defaultExpression->containsAnIntegral()) {
    return patTRUE ;
  }
  return patFALSE ;
}


patBoolean bioArithElem::containsASequence() const {
  for (map<patULong,bioExpression*>::const_iterator i = theDictionary.begin() ;
       i != theDictionary.end() ;
       ++i) {
    if (i->second->containsASequence()) {
      return patTRUE ;
    }
  }
  if (defaultExpression->containsASequence()) {
    return patTRUE ;
  }
  return patFALSE ;
}

void bioArithElem::generateGslFunctions(ostream& str_h,patString prefix, patError*& err) {
  for (map<patULong,bioExpression*>::const_iterator i = theDictionary.begin() ;
       i != theDictionary.end() ;
       ++i) {
    i->second->generateGslFunctions(str_h,prefix,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
  }
  defaultExpression->generateGslFunctions(str_h,prefix,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }

}
