//-*-c++-*------------------------------------------------------------
//
// File name : bioArithMult.cc
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Tue Apr 28 20:44:04 2009
//
//--------------------------------------------------------------------

#include <sstream>

#include "patDisplay.h"
#include "patErrNullPointer.h"

#include "bioArithMult.h"
#include "bioArithBinaryPlus.h"
#include "bioArithConstant.h"

bioArithMult::bioArithMult(bioExpression* par,
                                       bioExpression* left, 
                                       bioExpression* right) : bioExpression(par,left,right)
{
}

bioArithMult::~bioArithMult() {}

bioExpression::bioOperatorType bioArithMult::getOperatorType() const {
  return bioExpression::BINARY_OP ;
}

patString bioArithMult::getOperatorName() const {
  return ("*") ;
}

patReal bioArithMult::getValue(patError*& err) const  {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patReal();
  }
  
  patReal left = leftChild->getValue(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }
  patReal right = rightChild->getValue(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }
  patReal result = left * right ;

  return result ;
}

bioExpression* bioArithMult::getDerivative(patULong aLiteralId, 
					   patError*& err) const {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return NULL;
  }

  
  bioExpression* leftValue = leftChild->getDeepCopy(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  bioExpression* rightValue = rightChild->getDeepCopy(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  bioExpression* leftDeriv = leftChild->getDerivative(aLiteralId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  bioExpression* rightDeriv = rightChild->getDerivative(aLiteralId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }

  bioExpression* l = new bioArithMult(NULL,leftDeriv,rightValue) ;
  bioExpression* r = new bioArithMult(NULL,leftValue,rightDeriv) ;
  if (leftChild->dependsOf(aLiteralId)) {
    if (rightChild->dependsOf(aLiteralId)) {
      // Left variable, right variable
      bioExpression* result = new bioArithBinaryPlus(NULL,l,r) ;
      return result ;
    }
    else {
      // Left variable, right constant
      return l ;
    }
  }
  else {
    if (rightChild->dependsOf(aLiteralId)) {
      //Left constant, right variable
      return r ;
    }
    else {
      //Left constant, right constant
      bioExpression* result = new bioArithConstant(NULL,0) ;
      return result ;
    }
  }


}

bioArithMult* bioArithMult::getDeepCopy(patError*& err) {
  bioExpression* leftClone(NULL) ;
  bioExpression* rightClone(NULL) ;
  if (leftChild != NULL) {
    leftClone = leftChild->getDeepCopy(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    if (leftClone == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return NULL ;
    }
  }
  if (rightChild != NULL) {
    rightClone = rightChild->getDeepCopy(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    if (rightClone == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return NULL ;
    }
  }

  bioArithMult* theNode = 
    new bioArithMult(NULL,leftClone,rightClone) ;
  leftClone->setParent(theNode) ;
  rightClone->setParent(theNode) ;
  return theNode ;
}

patString bioArithMult::getCppCode(patString prefix, patError*& err) {
  if (codeMustBeGenerated) {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString();
  }
  
  patBoolean lSimple = leftChild->isSimple() ;
  patString vLeft ;
  if (lSimple) {
    vLeft = leftChild->getSimpleCppCode(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  else {
    vLeft = leftChild->getCppCode(prefix, err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  patBoolean rSimple = rightChild->isSimple() ;
  patString vRight ;
  if (rSimple) {
    vRight = rightChild->getSimpleCppCode(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  else {
    vRight = rightChild->getCppCode(prefix, err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  stringstream str ;
  str << getCppResultDeclaration(prefix) << ";" << endl ;
  if (!lSimple) {
    str << vLeft << endl ;
  }
  if (!rSimple) {
    str << vRight << endl ;
  }
  if (!lSimple) {
    str << "if (" <<leftChild->getCppResult(prefix) << "== 0) {" << endl ;
    str << "    " << getCppResult(prefix) << "= 0 ;" << endl ;
    str << "}" << endl ;
  }
  else {
    str << "if ((" << vLeft << ") == 0) {" << endl ;
    str << "    " << getCppResult(prefix) << "= 0 ;" << endl ;
    str << "}" << endl ;
  }
  if (!rSimple) {
    str << "else if (" <<rightChild->getCppResult(prefix) << "== 0) {" << endl ;
    str << "    " << getCppResult(prefix) << "= 0 ;" << endl ;
    str << "}" << endl ;
  }
  else {
    str << "else if ((" << vRight << ") == 0) {" << endl ;
    str << "    " << getCppResult(prefix) << "= 0 ;" << endl ;
    str << "}" << endl ;
  }
  str << " else {" << endl ;
  str << getCppResult(prefix) << "= (" ;
  if (lSimple) {
    str << vLeft ;
  }
  else {
    str << leftChild->getCppResult(prefix) ;
  }
  str << " ) * ( " ;
  if (rSimple) {
    str << vRight ;
  }
  else {
    str << rightChild->getCppResult(prefix) ;
  }
  str << ");" << endl ;
  str << "    if (!finite(" << getCppResult(prefix) << ")) {" << endl ;
  str << "      " << getCppResult(prefix) << " =  patMaxReal ;" << endl ;
  str << "    }" << endl ;

  str << "}" << endl ;
  //  str << "DEBUG_MESSAGE(\"" << getCppResult(prefix) << "= \" << " << getCppResult(prefix) << ");" << endl ;
  codeMustBeGenerated = patFALSE ;
  return patString(str.str()) ;
  }
  else {
    return patString() ;
  }

}


patString bioArithMult::getSimpleCppCode(patError*& err) {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString();
  }
  
  patString vLeft = leftChild->getSimpleCppCode(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  patString vRight = rightChild->getSimpleCppCode(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  stringstream str ;
  str << "(" << vLeft << ")*(" << vRight << ")"  ;
  return patString(str.str()) ;
}

patBoolean bioArithMult::isStructurallyZero() const {
  if (leftChild->isStructurallyZero()) {
    return patTRUE ;
  }
  if (rightChild->isStructurallyZero()) {
    return patTRUE ;
  }
  return patFALSE ;
}

patBoolean bioArithMult::isSimple() const {
  return patFALSE ;
}

patString bioArithMult::getExpressionString() const {
  stringstream str ;
  str << '*' ;
  if (leftChild != NULL) {
    str << '{' << leftChild->getExpressionString() << '}' ;
  }
  if (rightChild != NULL) {
    str << '{' << rightChild->getExpressionString() << '}' ;
  }
  return patString(str.str()) ;

}
