//-*-c++-*------------------------------------------------------------
//
// File name : bioArithNormalCdf.cc
// Author :    Michel Bierlaire
// Date :      Tue Jun  8 18:00:07 2010
//
//--------------------------------------------------------------------

#include <sstream>

#include "patDisplay.h"
#include "patErrNullPointer.h"
#include "patErrMiscError.h"
#include "bioArithNormalCdf.h"
#include "bioArithDivide.h"
#include "bioArithMult.h"
#include "bioArithConstant.h"
#include "bioArithExp.h"
#include "bioArithUnaryMinus.h"

bioArithNormalCdf::bioArithNormalCdf(bioExpression* par,
                         bioExpression* left) 
    : bioExpression(par,left,NULL)
{
}

bioArithNormalCdf::~bioArithNormalCdf() {}

bioExpression::bioOperatorType bioArithNormalCdf::getOperatorType() const {
  return bioExpression::UNARY_OP ;
}

patString bioArithNormalCdf::getOperatorName() const {
  return ("normalCdf") ;
}

patReal bioArithNormalCdf::getValue(patError*& err) const  {
  if (leftChild == NULL) {
    err = new patErrNullPointer("patArithNode") ;
    WARNING(err->describe()) ;
    return patReal();
  }
  
  patReal childresult = leftChild->getValue(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }

  patReal result = patNormalCdf::the()->compute(childresult,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }
  
  return result ;
}


bioExpression* bioArithNormalCdf::getDerivative(patULong aLiteralId, patError*& err) const {
  if (leftChild == NULL) {
    err = new patErrNullPointer("patArithNode") ;
    WARNING(err->describe()) ;
    return NULL ;
  }
  

  bioExpression* value1 = leftChild->getDeepCopy(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  bioExpression* value2 = leftChild->getDeepCopy(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }

  bioExpression* deriv = leftChild->getDerivative(aLiteralId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }

  // value * value 

  bioExpression* y2 = new bioArithMult(NULL,value1,value2) ;
  
  // (value * value) / 2

  bioExpression* two = new bioArithConstant(NULL,2.0) ;

  bioExpression* d = new bioArithDivide(NULL,y2,two) ;

  // - (value * value) / 2

  bioExpression* minusd = new bioArithUnaryMinus(NULL,d) ;

  // exp(- (value * value) / 2)

  bioExpression* expo = new bioArithExp(NULL,minusd) ;

  // oneDivSqrtTwoPi * exp(- (value * value) / 2)


  bioExpression*  oneDivSqrtTwoPi =  new bioArithConstant(NULL,1.0/sqrt(2*pi)) ;

  bioExpression* m = new bioArithMult(NULL,oneDivSqrtTwoPi,expo) ;

  bioExpression* result = new bioArithMult(NULL,m,deriv) ;

  return result ;
}


bioArithNormalCdf* bioArithNormalCdf::getDeepCopy(patError*& err) {
  bioExpression* leftClone(NULL) ;
  bioExpression* rightClone(NULL) ;
  if (leftChild != NULL) {
    leftClone = leftChild->getDeepCopy(err) ;
  }
  if (rightChild != NULL) { // Must be NULL
    rightClone = rightChild->getDeepCopy(err) ;
  }

  bioArithNormalCdf* theNode = 
    new bioArithNormalCdf(NULL,leftClone) ;
  leftClone->setParent(theNode) ;
  return theNode ;
}


patString bioArithNormalCdf::getCppCode(patString prefix, patError*& err) {
  if (codeMustBeGenerated) {
  if (leftChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString();
  }
  
  patString vLeft = leftChild->getCppCode(prefix, err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }


  stringstream str ;
  str << vLeft << endl ;
  str << getCppResultDeclaration(prefix) << " ;" << endl ;
  str << "   " << getCppResult(prefix) << " = patNormalCdf::the()->compute(" << leftChild->getCppResult(prefix) << ",err);" << endl ;
  str << "  if (err != NULL) {" << endl ;
  str << "    WARNING(err->describe()) ;" << endl ;
  str << "    return NULL;" << endl ;
  str << "  }" << endl ;
  
  //  str << "DEBUG_MESSAGE(\"" << getCppResult(prefix) << "= \" << " << getCppResult(prefix) << ");" << endl ;
  codeMustBeGenerated = patFALSE ;
  
  return patString(str.str()) ;
  }
  else {
    return patString() ;
  }
}

patString bioArithNormalCdf::getSimpleCppCode(patError*& err) {
  if (leftChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString();
  }
  patString vLeft = leftChild->getSimpleCppCode(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  stringstream str ;
  str << "patNormalCdf()(" << vLeft << ")" ;
  return patString(str.str()) ;
}

patBoolean bioArithNormalCdf::isSimple() const {
  return patFALSE ;
}

patString bioArithNormalCdf::getExpressionString() const {
  stringstream str ;
  str << '$NC' ;
  if (leftChild != NULL) {
    str << '{' << leftChild->getExpressionString() << '}' ;
  }
  return patString(str.str()) ;
}

