//-*-c++-*------------------------------------------------------------
//
// File name : bioArithPower.cc
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Tue May  5 12:05:09 2009
//
//--------------------------------------------------------------------

#include <sstream>

#include "patDisplay.h"
#include "patErrNullPointer.h"
#include "patErrMiscError.h"
#include "bioArithPower.h"
#include "bioArithConstant.h"
#include "bioArithLog.h"
#include "bioArithSequence.h"
#include "bioArithLiteral.h"
#include "bioArithMult.h"
#include "bioArithBinaryPlus.h"
#include "bioArithDivide.h"
#include "bioLiteralRepository.h"
#include "bioArithNotEqual.h"

bioArithPower::bioArithPower(bioExpression* par,
                             bioExpression* left,
                             bioExpression* right)
    : bioExpression(par,left,right)
{
}


bioArithPower::~bioArithPower() {}

bioExpression::bioOperatorType bioArithPower::getOperatorType() const {
  return bioExpression::BINARY_OP ;
}

patString bioArithPower::getOperatorName() const {
  return ("^") ;
}

patReal bioArithPower::getValue(patError*& err) const  {
  if (leftChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patReal();
  }
  
  patReal l = leftChild->getValue(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }
  if (l == 0) {
    return 0 ;
  }
  patReal r = rightChild->getValue(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }

  patReal result = pow(l,r) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }
  if (finite(result) == 0) {
    stringstream str ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patReal();
    }
    str << "Numerical problem in evaluating " << l << "^" << r << " within " << getInfo() ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return patReal() ;
  }
  
  return result ;
}


bioExpression* bioArithPower::getDerivative(patULong aLiteralId, 
					    patError*& err) const {

  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    WARNING(getInfo()) ;
    return NULL;
  }

  if (!dependsOf(aLiteralId) || leftChild->isStructurallyZero()) {
    bioExpression* result = new bioArithConstant(NULL,0.0) ;
    return result ;
  }
  
  bioExpression* leftValue = leftChild->getDeepCopy(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }

  bioExpression* rightValue = rightChild->getDeepCopy(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }


  bioExpression* leftDeriv = leftChild->getDerivative(aLiteralId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }


  bioExpression* rightDeriv = rightChild->getDerivative(aLiteralId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }


  bioExpression* power = new bioArithPower(NULL,leftValue,rightValue) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }

  //  Derivative = 
  // power * [rightValue * leftDeriv / leftValue + rightDeriv * ln(leftValue)]
  
  patBoolean both(patTRUE) ;
  bioExpression* term1(NULL) ;
  bioExpression* term2(NULL) ;
  if (leftChild->dependsOf(aLiteralId)) {
    bioExpression* n = new bioArithMult(NULL,rightValue,leftDeriv) ;
    term1 = new bioArithDivide(NULL,n,leftValue) ;
  }
  else {
    both = patFALSE ;
  }
  if (rightChild->dependsOf(aLiteralId)) {
    // If the leftValue is 0, no log can be taken
//     patULong theId = bioLiteralRepository::the()->addIntermediateValue(err) ;
//     if (err != NULL) {
//       WARNING(err->describe()) ;
//     }
//     bioArithLiteral* theSimplifiedExpression =  new bioArithLiteral(NULL,theId) ;
    bioExpression* l= new bioArithLog(NULL,leftValue) ;
    bioExpression* t2 = new bioArithMult(NULL,rightDeriv,l) ;
    bioExpression* zero = new bioArithConstant(NULL,0.0) ;
    bioExpression* cond = new bioArithNotEqual(NULL,leftValue,zero) ;
    term2 = new bioArithMult(NULL,cond,t2) ;
    //    term2 = new bioArithSequence(NULL,leftValue,tmp,theId,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
    }
  }
  else {
    both = patFALSE ;
  }

  bioExpression* sum ;
  if (both) {
    sum = new bioArithBinaryPlus(NULL,term1,term2) ;
  }
  else {
    if (term1 != NULL) {
      sum = term1 ;
    }
    else if (term2 != NULL) {
      sum = term2 ;
    }
    else {
      // Should never happen
      err = new patErrMiscError("Something's wrong... One of the two terms shoud be non null.");
      WARNING(err->describe()) ;
      return NULL ;
    }
  }

  bioExpression* result = new bioArithMult(NULL,power,sum) ;
  return result ;
}


bioArithPower* bioArithPower::getDeepCopy(patError*& err) {
  bioExpression* leftClone(NULL) ;
  bioExpression* rightClone(NULL) ;
  if (leftChild != NULL) {
    leftClone = leftChild->getDeepCopy(err) ;
  }
  if (rightChild != NULL) {
    rightClone = rightChild->getDeepCopy(err) ;
  }

  bioArithPower* theNode = 
    new bioArithPower(NULL,leftClone,rightClone) ;
  leftClone->setParent(theNode) ;
  rightClone->setParent(theNode) ;
  return theNode ;
}

patString bioArithPower::getSimpleCppCode(patError*& err) {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString();
  }
  
  patString vLeft = leftChild->getSimpleCppCode(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  patString vRight = rightChild->getSimpleCppCode(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  stringstream str ;
  str << "pow(" << vLeft << "," << vRight << ")" ;
  return patString(str.str()) ;
}

patString bioArithPower::getCppCode(patString prefix, patError*& err) {
  if (codeMustBeGenerated) {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString();
  }
  
  patBoolean lSimple = leftChild->isSimple() ;
  patString vLeft ;
  if (lSimple) {
    vLeft = leftChild->getSimpleCppCode(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  else {
    vLeft = leftChild->getCppCode(prefix, err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  patBoolean rSimple = rightChild->isSimple() ;
  patString vRight ;
  if (rSimple) {
    vRight = rightChild->getSimpleCppCode(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  else {
    vRight = rightChild->getCppCode(prefix, err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  stringstream str ;
  if (!lSimple) {
    str << vLeft << endl ;
  }
  if (!rSimple) {
    str << vRight << endl ;
  }
  str << getCppResultDeclaration(prefix) << "=pow(" ;
  if (lSimple) {
    str << vLeft ;
  }
  else {
    str << leftChild->getCppResult(prefix) ;
  }
  str << "," ;
  if (rSimple) {
    str << vRight ;
  }
  else {
    str << rightChild->getCppResult(prefix) ;
  }
  str << ");" << endl ;
  //  str << "DEBUG_MESSAGE(\"" << getCppResult(prefix) << "= \" << " << getCppResult(prefix) << ");" << endl ;
  codeMustBeGenerated = patFALSE ;
  return patString(str.str()) ;
  }
  else {
    return patString() ; 
  }
}

patBoolean bioArithPower::isStructurallyZero() const {
  return (leftChild->isStructurallyZero()) ;
}

patString bioArithPower::getExpressionString() const {
  stringstream str ;
  str << '^' ;
  if (leftChild != NULL) {
    str << '{' << leftChild->getExpressionString() << '}' ;
  }
  if (rightChild != NULL) {
    str << '{' << rightChild->getExpressionString() << '}' ;
  }
  return patString(str.str()) ;

}
