//-*-c++-*------------------------------------------------------------
//
// File name : bioArithSequence.cc
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Mon Aug 31 14:55:50 2009
//
//--------------------------------------------------------------------

#include "bioArithSequence.h"
#include "bioArithConstant.h"
#include "bioArithLiteral.h"
#include "bioArithMult.h"
#include "bioArithBinaryPlus.h"
#include "bioLiteralRepository.h"
#include "patErrNullPointer.h"
#include "patErrMiscError.h"
#include "patDisplay.h"
#include "patMath.h"

bioArithSequence::bioArithSequence(bioExpression* par, 
				   bioExpression* left, 
				   bioExpression* right,
				   patULong id, 
				   patError*& err)  :
  bioExpression(par,left,right), theId(id) {


}


bioArithSequence::bioArithSequence(bioExpression* par,
				   bioExpression* left, 
				   bioExpression* right,
				   patError*& err) :
  bioExpression(par,left,right){

  theId = bioLiteralRepository::the()->addIntermediateValue(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
  }
  bioArithLiteral* theSimplifiedExpression =  new bioArithLiteral(NULL,theId) ;

  //  DEBUG_MESSAGE("--> right before = " << *rightChild) ;

  rightChild->replaceExpression(left,theSimplifiedExpression,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
  }

//   DEBUG_MESSAGE("--> right after = " << *rightChild) ;
//   DEBUG_MESSAGE("--> sequence: " << *this) ;
}

bioArithSequence::~bioArithSequence() {

}

bioExpression::bioOperatorType bioArithSequence::getOperatorType() const {
  return bioExpression::SEQUENCE_OP ;
}

patString bioArithSequence::getOperatorName() const {
  return patString("Sequence") ;
}

patReal bioArithSequence::getValue(patError*& err) const {
  err = new patErrMiscError("Cannot compute the value of a sequence") ;
  WARNING(err->describe()) ;
  return patReal() ;
}

bioExpression* bioArithSequence::getDerivative(patULong aLiteralId, 
					       patError*& err) const {

  patBoolean leftDependsOnBeta = leftChild->dependsOf(aLiteralId) ;
  bioExpression* leftDeriv(NULL) ;
  if (leftDependsOnBeta) {
    leftDeriv = leftChild->getDerivative(aLiteralId,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
  }

  patBoolean rightDependsOnBeta = rightChild->dependsOf(aLiteralId) ;
  bioExpression* rightDeriv(NULL) ;
  if (rightDependsOnBeta) {
    rightDeriv = rightChild->getDerivative(aLiteralId,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
  }
  
  patBoolean rightDependsOnIntermediate = rightChild->dependsOf(theId) ;
  bioExpression* derivWrtIntermediateVariable(NULL) ;
  if (rightDependsOnIntermediate) {
    derivWrtIntermediateVariable = rightChild->getDerivative(theId,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
  }

  if (!leftDependsOnBeta || !rightDependsOnIntermediate) {
    if (rightDependsOnBeta) {
      bioExpression* leftClone = leftChild->getDeepCopy(err) ;

      bioArithSequence* theSeqDeriv = new bioArithSequence(NULL,leftClone,rightDeriv,theId,err) ;
      return theSeqDeriv ;
    }
    else {
      bioExpression* result = new bioArithConstant(NULL,0.0) ;
      return result ;
    }
  }
  bioExpression* term2 = new bioArithMult(NULL, leftDeriv,derivWrtIntermediateVariable) ;
  if (rightDependsOnBeta) {
    bioExpression* theDeriv = new bioArithBinaryPlus(NULL,rightDeriv,term2) ;
    bioExpression* leftClone = leftChild->getDeepCopy(err) ;

    bioArithSequence* theSeqDeriv = new bioArithSequence(NULL,leftClone,theDeriv,theId,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    return theSeqDeriv ;
  }
  else {
    bioExpression* leftClone = leftChild->getDeepCopy(err) ;
    bioArithSequence* theSeqDeriv = new bioArithSequence(NULL,leftClone,term2,theId,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    return theSeqDeriv ;
  }
  err = new patErrMiscError("Should not reach this point as all cases are already covered") ;
  WARNING(err->describe()) ;
  return NULL ;
}

bioExpression* bioArithSequence::getDeepCopy(patError*& err) {
  bioExpression* leftClone(NULL) ;
  bioExpression* rightClone(NULL) ;
  if (leftChild != NULL) {
    leftClone = leftChild->getDeepCopy(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    if (leftClone == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return NULL ;
    }
  }
  if (rightChild != NULL) {
    rightClone = rightChild->getDeepCopy(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    if (rightClone == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return NULL ;
    }
  }

  bioArithSequence* result = new bioArithSequence(NULL,
						  leftClone,
						  rightClone,
						  theId,
						  err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }

  leftClone->setParent(result) ;
  rightClone->setParent(result) ;
  return result ;
}

patString bioArithSequence::getCppCode(patString prefix, patError*& err) {
  if (codeMustBeGenerated) {
    stringstream str ;

    str << "// =============================" << endl ;
    str << "// First expression of sequence " << theId << endl ;
    str << "// " << *leftChild << endl ;
    str << "// =============================" << endl ;
    if (!leftChild->isSimple()) {
      str << leftChild->getCppCode(prefix, err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return patString();
      }
    }
    str << getCppResultDeclaration(prefix) << "=" ;
    if (leftChild->isSimple()) {
      str <<  leftChild->getSimpleCppCode(err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return patString();
      }
    }
    else {
      str << leftChild->getCppResult(prefix) ;
    }
    str << " ; " << endl ; 
    //    str << "DEBUG_MESSAGE(\"First in sequence: " << getCppResult(prefix) << "= \" << " << getCppResult(prefix) << ");" << endl ;
    patString l = bioLiteralRepository::the()->getCppCode(theId,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
    if (!bioLiteralRepository::the()->isFlagSet(theId)) {
      str << "patReal " ;
      bioLiteralRepository::the()->setFlag(theId) ;
    }
    str <<  l << " = " << getCppResult(prefix) << ";" << endl ;
    
    str << "// =============================" << endl ;
    str << "// Second expression of sequence " << theId << endl ;
    str << "// " << *rightChild << endl ;
    str << "// =============================" << endl ;
    str << rightChild->getCppCode(prefix, err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
    str << getCppResult(prefix) << " = " << rightChild->getCppResult(prefix) << " ;"<< endl ;
    //    str << "DEBUG_MESSAGE(\"Second in sequence: " << rightChild->getCppResult(prefix) << "= \" << " << getCppResult(prefix) << ");" << endl ;
    return patString(str.str()) ;
    codeMustBeGenerated = patFALSE ;
  }
  else {
    return patString() ;
  }
}

patString bioArithSequence::getSimpleCppCode(patError*& err) {
  err = new patErrMiscError("No simple c++ code for a sequence") ;
  WARNING(err->describe());
  return patString() ;
}

patBoolean bioArithSequence::isSimple() const {
  return patFALSE ;
}



patString bioArithSequence::getExpression(patError*& err) const {
  stringstream str ;
  if (leftChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString() ;
  }
  patString l = leftChild->getExpression(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString() ;
  }
  if (rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString() ;
  }
  patString r = rightChild->getExpression(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString() ;
  }
  patString theName = bioLiteralRepository::the()->getName(theId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString() ;
  }
  str << "Replace <" << theName << "> by " << l << " in <" << r << ">" ;
  return patString(str.str()) ;
}

patULong bioArithSequence::getNumberOfOperations() const {
  patULong result = 0 ;
  if (leftChild != NULL) {
    result += leftChild->getNumberOfOperations() ;
  }
  if (rightChild != NULL) {
    result += rightChild->getNumberOfOperations() ;
  }
  return result ;

}



patString bioArithSequence::getExpressionString() const {
  stringstream str ;
  str << '$Q' ;
  if (leftChild != NULL) {
    str << '{' << leftChild->getExpressionString() << '}' ;
  }
  if (rightChild != NULL) {
    str << '{' << rightChild->getExpressionString() << '}' ;
  }
  return patString(str.str()) ;
  
}

patULong bioArithSequence::countExpressions(map<patString,patULong>* theExpressions, 
					 map<patString,list<patULong> >* theIds, 
					map<patString,list<bioExpression*> >* thePointers,
					 patError*& err)   {

  // THe only difference with another expression is that the sequence it self is not considered, only its children.

  if (containsAnIterator()) {
    stringstream str ;
    str << "Expression to simplify cannot contain an iterator: " << *this;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return 0;
  }

  patULong result(0) ;
  if (leftChild != NULL) {
    result += leftChild->countExpressions(theExpressions,theIds,thePointers,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patULong() ;
    }
  }
  if (rightChild != NULL) {
    result += rightChild->countExpressions(theExpressions,theIds,thePointers,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patULong() ;
    }
  }
  return result ;
}

patBoolean bioArithSequence::isSequence() const {
  return patTRUE ;
}

void bioArithSequence::simplifyLowerIteratorOrSequence(patError*& err) {
  if (leftChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return ;
  }
  if (leftChild->canBeSimplified()) {
    bioExpression* newExpression = simplify(leftChild,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
    leftChild = newExpression ;
    leftChild->setParent(this) ;
  }
  if (rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return ;
  }
  if (rightChild->canBeSimplified()) {
    bioExpression* newExpression = simplify(rightChild,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
    rightChild = newExpression ;
    rightChild->setParent(this) ;
  }
}
