from biogeme import *
from mev import *

# Implements the nested logit model directly from the probability
# formula. It is the most efficient function, computationally
# speaking.
#
# Nests are defined as follows:
#nesta = MUA , [1,2,3]
#nestb = MUB , [4,5,6]
#nests = nesta, nestb
#
def nested(V,availability,nests,choice) :

    sumdenoms = 0.0
    musPerAlt = {}
    allDenomsPerAlt = {}
    chosen = Elem(V,choice)
    for m in nests:
        denom = 0.0
        for i in m[1]:
            denom += availability[i] * exp((V[i]-chosen) * m[0])

        # Store nest information per alternative to retrieve them
        # easily using Elem
        for i in m[1]:
            allDenomsPerAlt[i] = denom
            musPerAlt[i] = m[0]
        sumdenoms += denom ** (1.0/m[0])

    power = ((1.0 / Elem(musPerAlt,choice))-1.0)
    print("Power=",power) ;
    P = Elem(availability,choice) * Elem(allDenomsPerAlt,choice) ** power / sumdenoms 
    return P

# This version is less performant. It is kept for debuggin purposes.
def nestedold(V,availability,nests,choice) :

    sumdenoms = 0.0
    musPerAlt = {}
    allDenomsPerAlt = {}
    chosen = Elem(V,choice)
    for m in nests:
        denom = 0.0
        for i in m[1]:
            denom += availability[i] * exp(V[i] * m[0])

        # Store nest information per alternative to retrieve them
        # easily using Elem
        for i in m[1]:
            allDenomsPerAlt[i] = denom
            musPerAlt[i] = m[0]
        sumdenoms += denom ** (1.0/m[0])

    power = ((1.0 / Elem(musPerAlt,choice))-1.0)
    print("Power=",power) ;
    P = Elem(availability,choice) * exp(Elem(musPerAlt,choice) * Elem(V,choice)) * Elem(allDenomsPerAlt,choice) ** power / sumdenoms 
    return P




#Same as above, without availability
def nestedFull(V,nests,choice) :

    sumdenoms = 0.0
    musPerAlt = {}
    allDenomsPerAlt = {}
    for m in nests:
        denom = 0.0
        for i in m[1]:
            denom += exp(V[i] * m[0])

        # Store nest information per alternative to retrieve them
        # easily using Elem
        for i in m[1]:
            allDenomsPerAlt[i] = denom
            musPerAlt[i] = m[0]
        sumdenoms += denom ** (1.0/m[0])

    power = ((1.0 / Elem(musPerAlt,choice))-1.0)
    print("Power=",power) ;
    P = exp(Elem(musPerAlt,choice) * Elem(V,choice)) * Elem(allDenomsPerAlt,choice) ** power / sumdenoms 
    return P

# In this implementation, all utilities are shifted by V_i. It makes
# the model more robust numerically, but involves more operations.
def shiftedNested(V,availability,nests,choice) :

    chosen = Elem(V,choice)
    sumdenoms = 0.0
    musPerAlt = {}
    allDenomsPerAlt = {}
    for m in nests:
        denom = 0.0
        for i in m[1]:
            denom += availability[i] * exp((V[i]-chosen) * m[0])

        # Store nest information per alternative to retrieve them
        # easily using Elem
        for i in m[1]:
            allDenomsPerAlt[i] = denom
            musPerAlt[i] = m[0]
        sumdenoms += denom ** (1.0/m[0])

    P = Elem(availability,choice) * Elem(allDenomsPerAlt,choice) ** ((1.0/Elem(musPerAlt,choice))-1.0) / sumdenoms 
    return P

# Same as above, all alternatives are always available
def shiftedNestedFull(V,nests,choice) :

    chosen = Elem(V,choice)
    sumdenoms = 0.0
    musPerAlt = {}
    allDenomsPerAlt = {}
    for m in nests:
        denom = 0.0
        for i in m[1]:
            denom += exp((V[i]-chosen) * m[0])

        # Store nest information per alternative to retrieve them
        # easily using Elem
        for i in m[1]:
            allDenomsPerAlt[i] = denom
            musPerAlt[i] = m[0]
        sumdenoms += denom ** (1.0/m[0])

    P = Elem(allDenomsPerAlt,choice) ** ((1.0/Elem(musPerAlt,choice))-1.0) / sumdenoms 
    return P

# Implements the nested logit model as a MEV model. 
def getMevForNested(V,availability,nests) :

    y = {}
    for i,v in V.items() :
        y[i] = exp(v)
    
    Gi = {}
    for m in nests:
        sum = 0.0
        for i in m[1]:
            sum += availability[i] * y[i] ** m[0]
        for i in m[1]:
            Gi[i] = availability[i] * y[i]**(m[0]-1.0) * sum ** (1.0/m[0] - 1.0)
    return Gi


def getMevShiftedForNested(V,availability,nests,choice) :

    chosen = Elem(V,choice)
    y = {}
    for i,v in V.items() :
        y[i] = exp(v-chosen)
    
    Gi = {}
    for m in nests:
        sum = 0.0
        for i in m[1]:
            sum += availability[i] * y[i] ** m[0]
        for i in m[1]:
            Gi[i] = availability[i] * y[i]**(m[0]-1.0) * sum ** (1.0/m[0] - 1.0)
    return Gi



# Implements the nested logit model as a MEV model. 
def nestedMev(V,availability,nests,choice) :

    Gi = getMevShiftedForNested(V,availability,nests,choice)
    P = mev_shifted(V,Gi,choice) 
    return P

# Implements the nested logit model as a MEV model. mu is also a parameter here
def nestedMevMu(V,availability,nests,choice,mu) :

    chosen = Elem(V,choice)
    y = {}
    for i,v in V.items() :
        y[i] = exp(v-chosen)
    
    Gi = {}
    for m in nests:
        sum = 0.0
        for i in m[1]:
            sum += availability[i] * y[i] ** m[0]
        for i in m[1]:
            Gi[i] = availability[i] * mu * y[i]**(m[0]-1.0) * sum ** ((mu/m[0]) - 1.0)
    
    P = mev_shifted(V,Gi,choice) 
    return P
